/*
 * Succinct representation of RMQ information
 * by Johannes Fischer, additional changes by Niko Vlimki.
 * Used with Fischer's permission.
 *
 * Johannes Fischer and Volker Heun, A New succinct representation 
 * of RMQ information and improvements in the Enhanced suffix array.
 * ESCAPE'07, 2007.
 */

#ifndef _RMQ_succinct_hpp_
#define _RMQ_succinct_hpp_

//#define MEM_COUNT

#include <cstdlib>
#include <climits>
#include <iostream>
#include <cmath>
#include "CHgtArray.h"

typedef int DT;                 // use long for 64bit-version (but take care of fast log!)
typedef unsigned int DTidx;     // for indexing in arrays
typedef unsigned char DTsucc;
typedef unsigned short DTsucc2;

class RMQ_succinct {
public:
	// liefert RMQ[i,j]
	DTidx query(DTidx, DTidx);

	RMQ_succinct(CHgtArray *, DTidx n);

	~RMQ_succinct();

    int GetC(ulong i) const
    {
        if (i == 0 || i >= n) // FIXME
            return 0;
        return hgtA->GetPos(i) + 1;
    }
    
protected:
	// array
	CHgtArray *hgtA;

	// size of array a
	DTidx n;

	// table M for the out-of-block queries (contains indices of block-minima)
	DTsucc** M;


    
	// because M just stores offsets (rel. to start of block), this method
	// re-calculates the true index:
	inline DTidx m(DTidx k, DTidx block) { return M[k][block]+(block*sprime); }

	// depth of table M:
	DTidx M_depth;

	// table M' for superblock-queries (contains indices of block-minima)
	DTidx** Mprime;

	// depth of table M':
	DTidx Mprime_depth;

	// type of blocks
	DTsucc2 *type;

	// precomputed in-block queries
	DTsucc** Prec;

	// microblock size
	DTidx s;

	// block size
	DTidx sprime;

	// superblock size
	DTidx sprimeprime;

	// number of blocks (always n/sprime)
	DTidx nb;

	// number of superblocks (always n/sprimeprime)
	DTidx nsb;

	// number of microblocks (always n/s)
	DTidx nmb;

	// return microblock-number of entry i:
	inline DTidx microblock(DTidx i) { return i/s; }

	// return block-number of entry i:
	inline DTidx block(DTidx i) { return i/sprime; }

	// return superblock-number of entry i:
	inline DTidx superblock(DTidx i) { return i/sprimeprime; }

	// precomputed Catalan triangle (17 is enough for 64bit computing):
	static const DTidx Catalan[17][17];

	// minus infinity (change for 64bit version)
	static const DT minus_infinity;

 	// stuff for clearing the least significant x bits (change for 64-bit computing)
	static const DTsucc HighestBitsSet[8];
	DTsucc clearbits(DTsucc, DTidx);

	// Least Significant Bits for 8-bit-numbers:
	static const char LSBTable256[256];

    // return least signigicant bit in constant time (change for 64bit version)
	DTidx lsb(DTsucc);

	// the following stuff is for fast base 2 logarithms:
	// (currently only implemented for 32 bit numbers)
	static const char LogTable256[256];

	DTidx log2fast(DTidx);
};

#endif
