/***************************************************************************
 *   Copyright (C) 2007 by Niko Vlimki and Johannes Fischer              *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#include "FrequentStrings.h"

int main(int argc, char *argv[]) {
    char *file1, *file2;
    if (argc != 7) {
        cout << "Usage:\n\n";
        cout << argv[0] << " [type] [database 1] [database 2] [query] [min] [max]\n";
        cout << "\nwhere [type] is either \"1\" (for raw text) or \"2\" (for FASTA format),\n";
        cout << "\nQuery type [query] is 'F' for Frequent string mining or 'E' for Emerging Substrings:\n";
        cout << "  [min] <= 1 is the minimum frequency threshold for [database 1],\n";
        cout << "  and [max] <= 1 is the maximum frequency threshold for [database 2].\nOR\n";
        cout << "  1/|[database 1]| <= [min] <= 1.0 is the support threshold for [database 1],\n";
        cout << "  and [max] >= 1 is the growth rate from [database 1] to [database 2].\n";
        return 0;
    }

    if (argv[4][0] != 'E' && argv[4][0] != 'F'){
        cerr << "Query type [query] should be 'E' or 'F'. Exit.\n";
        exit(0);
    }

    // combined minfreq and maxfreq
    if (argv[4][0] == 'F' && atof(argv[5]) > 1) {
        cerr << "Minimum frequency threshold must be less or equal to 1. Exit.\n";
        exit(0);
    }
    if (argv[4][0] == 'F' && atof(argv[6]) > 1) {
        cerr << "Maximum frequency threshold must be less or equal to 1. Exit.\n";
        exit(0);
    }

    // combined minfreq and maxfreq
    if (argv[4][0] == 'E' && (atof(argv[5]) <= 0.0 || atof(argv[5]) > 1.0)) {
        cerr << "Support threshold must be greater than 0 and less than or equal to 1. Exit.\n";
        exit(0);
    }
    if (argv[4][0] == 'E' && atof(argv[6]) < 1.0) {
        cerr << "Growth rate must not be less than 1. Exit.\n";
        exit(0);
    }

    file1 = argv[2];
    file2 = argv[3];
    FrequentStrings freq;
    freq.frequent(file1, file2, atoi(argv[1]), argv[4][0], atof(argv[5]), atof(argv[6]));
    return 0;
}
